/*
#    xfrmcalc - Scientific calculator based on Forms.
#    Copyright (C) 1996-1997  Martin Bartlett
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "forms.h"
#include "calculator.h"
#include "divide.xbm"
#include "decimal.xbm"
#include "comma.xbm"
#include "power.xbm"
#include "square.xbm"

char *build_constants(void);
char * build_ext_funcs(void);

FD_calculator *create_form_calculator(void)
{
  char	xbm_name[1024], *menu_items;
  int	next_start;

  FL_OBJECT *obj;
  FL_FORM   *form;
  FD_calculator *fdui = (FD_calculator *) fl_calloc(1, sizeof(FD_calculator));

  fdui->calculator = form = fl_bgn_form(FL_NO_BOX, 318, 236);
  fl_register_raw_callback(form, ButtonPressMask | ButtonReleaseMask,
		raw_calc_callback);
  fl_add_box(FL_UP_BOX,0,0,318,26,"");
  fl_add_box(FL_FLAT_BOX,0,25,318,211,"");


/*
 * MENU
 */

  fl_setpup_default_fontsize(FL_NORMAL_SIZE);
  fl_setpup_default_fontstyle(FL_BOLD_STYLE);

  obj = fl_add_menu(FL_PULLDOWN_MENU,5,2,28,21,"File");
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
    fl_set_menu(obj, "Exit");
    fl_set_object_callback(obj,ExitProgram,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,35,2,30,21,"Edit");
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
    fl_set_menu(obj, "Select|Paste");
    fl_set_object_callback(obj,CutPasteDisplay,0);

  next_start = 68;

  if (menu_items = build_constants())
  {
  	obj = fl_add_menu(FL_PULLDOWN_MENU,next_start,2,65,21,"Constants");
    	fl_set_object_boxtype(obj,FL_FLAT_BOX);
    	fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    	fl_set_object_lstyle(obj,FL_BOLD_STYLE);
    	fl_set_menu(obj, menu_items);
		free(menu_items);
    	fl_set_object_callback(obj,InsertConstant,0);
		next_start += 68;
  }

  if (menu_items = build_ext_funcs())
  {
  	obj = fl_add_menu(FL_PULLDOWN_MENU,next_start,2,63,21,"Functions");
    	fl_set_object_boxtype(obj,FL_FLAT_BOX);
    	fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    	fl_set_object_lstyle(obj,FL_BOLD_STYLE);
    	fl_set_menu(obj, menu_items);
		free(menu_items);
    	fl_set_object_callback(obj,EnterFunction,0);
  }

  obj = fl_add_menu(FL_PULLDOWN_MENU,280,2,30,21,"Help");
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
    fl_set_menu(obj, "Help...");
    fl_set_object_callback(obj,CalcHelp,0);

/*
 * BASE RADIO BUTTONS
 */
  fdui->BaseButtons = fl_bgn_group();
  obj = fl_add_frame(FL_ENGRAVED_FRAME,4,65,170,27,"");
  obj = fl_add_box(FL_FLAT_BOX,9,59,30,10,"Base");
  fdui->CheckHex = obj = fl_add_checkbutton(FL_RADIO_BUTTON,4,65,20,27,"Hex");
    fl_set_button_shortcut(obj,"#x",0);
    fl_set_object_callback(obj,SetHex,16);
  fdui->DecSwitch = obj = fl_add_checkbutton(FL_RADIO_BUTTON,46,65,20,27,"Dec");
    fl_set_button_shortcut(obj,"#d",0);
    fl_set_object_callback(obj,SetDec,10);
  fdui->OctSwitch = obj = fl_add_checkbutton(FL_RADIO_BUTTON,88,65,20,27,"Oct");
    fl_set_button_shortcut(obj,"#o",0);
    fl_set_object_callback(obj,SetOct,8);
  fdui->BinSwitch = obj = fl_add_checkbutton(FL_RADIO_BUTTON,130,65,20,27,"Bin");
    fl_set_button_shortcut(obj,"#b",0);
    fl_set_object_callback(obj,SetBin,2);
  fl_end_group();

/*
 * Dummy button to facilitate cycling through the bases
 */
  obj = fl_add_button(FL_HIDDEN_BUTTON, 40, 59, 2, 2, "");
	fl_set_object_callback(obj,CycleBase,0);
	fl_set_object_shortcut(obj,"	",0);

/*
 * ANGLE REPRESENTATION RADIO BUTTONS
 */
  fdui->AngleButtons = fl_bgn_group();
  obj = fl_add_frame(FL_ENGRAVED_FRAME,179,65,135,27,"");
  obj = fl_add_box(FL_FLAT_BOX,184,59,35,10,"Angle");
  fdui->DegSwitch = obj = fl_add_checkbutton(FL_RADIO_BUTTON,179,65,20,27,"Deg");
    fl_set_button_shortcut(obj,"^d",0);
    fl_set_object_callback(obj,SetDeg,0);
  fdui->RadSwitch = obj = fl_add_checkbutton(FL_RADIO_BUTTON,225,65,20,27,"Rad");
    fl_set_button_shortcut(obj,"^r",0);
    fl_set_object_callback(obj,SetRad,0);
  fdui->GraSwitch = obj = fl_add_checkbutton(FL_RADIO_BUTTON,270,65,20,27,"Gra");    fl_set_button_shortcut(obj,"^g",0);
    fl_set_object_callback(obj,SetGra,0);
  fl_end_group();

/*
 * Dummy button to facilitate cycling through the angles
 */
  obj = fl_add_button(FL_HIDDEN_BUTTON, 43, 62, 2, 2, "");
	fl_set_object_callback(obj,CycleAngle,0);
	fl_set_object_shortcutkey(obj,XK_ISO_Left_Tab);

/*
 * CLEAR KEYS
 */
  fdui->ClearAll = obj = fl_add_button(FL_NORMAL_BUTTON,284,97,30,23,"AC");
    fl_set_object_color(obj, FL_DARKTOMATO, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"^[",0);
    fl_set_object_callback(obj,ClearAll,0);
  fdui->CLear = obj = fl_add_button(FL_NORMAL_BUTTON,249,97,30,23,"C");
    fl_set_object_color(obj, FL_DARKTOMATO, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"#C",0);
    fl_set_object_callback(obj,Clear,0);
  obj = fl_add_button(FL_HIDDEN_BUTTON,280,97,3,23,"");
    fl_set_button_shortcut(obj,"^H",0);
    fl_set_object_callback(obj,Backspace,0);

/*
 * 2-OPERAND STACKING FUNCTIONS
 */
  fdui->Multiply = obj = fl_add_button(FL_NORMAL_BUTTON,214,153,30,23,"X");
    fl_set_button_shortcut(obj,"*xX",0);
    fl_set_object_callback(obj,EnterStackFunction,8);
  fdui->Divide = obj = fl_add_bitmapbutton(FL_NORMAL_BUTTON,249,153,30,23,"");
    fl_set_bitmapbutton_data(obj, divide_width, divide_height, divide_bits);
    fl_set_button_shortcut(obj,"/",0);
    fl_set_object_callback(obj,EnterStackFunction,9);
  fdui->Plus = obj = fl_add_button(FL_NORMAL_BUTTON,214,181,30,23,"+");
    fl_set_button_shortcut(obj,"+",0);
    fl_set_object_callback(obj,EnterStackFunction,6);
  fdui->Minus = obj = fl_add_button(FL_NORMAL_BUTTON,249,181,30,23,"-");
    fl_set_button_shortcut(obj,"-",0);
    fl_set_object_callback(obj,EnterStackFunction,7);
  fdui->And = obj = fl_add_button(FL_NORMAL_BUTTON,284,125,30,23,"&");
    fl_set_button_shortcut(obj,"&&",0);
    fl_set_object_callback(obj,EnterStackFunction,3);

  /* Hidden button implements shortcut for Mouse Button 3 */
  fdui->Xor = obj = fl_add_button(FL_HIDDEN_BUTTON,284,153,30,23,"");
    fl_set_button_shortcut(obj,"#\\^",0);
    fl_set_object_callback(obj,EnterStackFunction,2);
  fdui->Or = obj = fl_add_button(FL_NORMAL_BUTTON,284,153,30,23,"|  ^");
    fl_set_button_shortcut(obj,"|\\",0);
    fl_set_object_callback(obj,EnterStackFunction,1);

  /* Hidden button implements shortcut for Mouse Button 3 */
  fdui->Rsh = obj = fl_add_button(FL_HIDDEN_BUTTON,284,181,30,23,"");
    fl_set_button_shortcut(obj,">",0);
    fl_set_object_callback(obj,EnterStackFunction,5);
  fdui->Lsh = obj = fl_add_button(FL_NORMAL_BUTTON,284,181,30,23,"<  >");
    fl_set_button_shortcut(obj,"<",0);
    fl_set_object_callback(obj,EnterStackFunction,4);

  /* Hidden button implements shortcut for Mouse Button 3 */
  fdui->InvMod = obj = fl_add_button(FL_HIDDEN_BUTTON,284,209,30,23,"");
    fl_set_button_shortcut(obj,"#;",0);
    fl_set_object_callback(obj,EnterStackFunction,13);
  fdui->Mod = obj = fl_add_button(FL_NORMAL_BUTTON,284,209,30,23,"Mod");
    fl_set_button_shortcut(obj,";",0);
    fl_set_object_callback(obj,EnterStackFunction,10);

  /* Hidden button implements shortcut for Mouse Button 3 */
  fdui->InvPower = obj = fl_add_button(FL_HIDDEN_BUTTON,249,209,30,23,"");
    fl_set_button_shortcut(obj,"#p",0);
    fl_set_object_callback(obj,EnterStackFunction,12);
  fdui->Power = obj = fl_add_bitmapbutton(FL_NORMAL_BUTTON,249,209,30,23,"");
    fl_set_button_shortcut(obj,"pP",0);
    fl_set_bitmapbutton_data(obj, power_width, power_height, power_bits);
    fl_set_object_callback(obj,EnterStackFunction,11);

  fdui->Equals = obj = fl_add_button(FL_NORMAL_BUTTON,179,209,30,23,"=");
    fl_set_button_shortcut(obj,"=^M",0);
    fl_set_object_callback(obj,EnterEqual,0);

/*
 * MEMORY FUNCTIONS
 */
  fdui->MemoryRecall = obj = fl_add_button(FL_NORMAL_BUTTON,109,97,30,23,"Min");
    fl_set_object_callback(obj,Memory,3);
    fl_set_button_shortcut(obj,"&A",0);

  fdui->MemoryRecall = obj = fl_add_button(FL_NORMAL_BUTTON,144,97,30,23,"MR");
    fl_set_object_callback(obj,Memory,1);
    fl_set_button_shortcut(obj,"&B",0);
  obj = fl_add_button(FL_HIDDEN_BUTTON,179,97,30,23,""); 
   fl_set_object_callback(obj,Memory,2);
    fl_set_button_shortcut(obj,"&D",0);
  fdui->MemoryPlus = obj = fl_add_button(FL_NORMAL_BUTTON,179,97,30,23,"M+-"); 
   fl_set_object_callback(obj,Memory,2);
    fl_set_button_shortcut(obj,"&C",0);
  fdui->MemoryClear = obj = fl_add_button(FL_NORMAL_BUTTON,214,97,30,23,"MC"); 
   fl_set_object_callback(obj,Memory,0);
    fl_set_button_shortcut(obj,"#&B",0);


/*
 * PARENTHESES
 */
  fdui->OpenParen = obj = fl_add_button(FL_NORMAL_BUTTON,214,125,30,23,"((...");
    fl_set_button_shortcut(obj,"(",0);
    fl_set_object_callback(obj,EnterOpenParen,0);
  fdui->CloseParen = obj = fl_add_button(FL_NORMAL_BUTTON,249,125,30,23,"...))");
    fl_set_button_shortcut(obj,")",0);
    fl_set_object_callback(obj,EnterCloseParen,0);

/*
 * 1-OPERAND NON-STACKING FUNCTIONS
 */
  /* Hidden button implements shortcut for Mouse Button 3 */
  fdui->Frac = obj = fl_add_button(FL_HIDDEN_BUTTON,39,97,30,20,"");
    fl_set_button_shortcut(obj,"#i",0);
    fl_set_object_callback(obj,EnterInt,1);
  fdui->Int = obj = fl_add_button(FL_NORMAL_BUTTON,39,97,30,20,"Int");
    fl_set_button_shortcut(obj,"iI",0);
    fl_set_object_callback(obj,EnterInt,0);

  fdui->Negate = obj = fl_add_button(FL_NORMAL_BUTTON,39,120,30,20,"+/-");
    fl_set_button_shortcut(obj,"#-",0);
    fl_set_object_callback(obj,EnterNegate,0);
  fdui->Recip = obj = fl_add_button(FL_NORMAL_BUTTON,39,143,30,20,"1/x");
    fl_set_button_shortcut(obj,"'",0);
    fl_set_object_callback(obj,EnterRecip,0);
  fdui->Factorial = obj = fl_add_button(FL_NORMAL_BUTTON,39,166,30,20,"x!");
    fl_set_button_shortcut(obj,"!",0);
    fl_set_object_callback(obj,EnterFactorial,0);

  /* Hidden button implements shortcut for Mouse Button 3 */
  fdui->SqRoot = obj = fl_add_button(FL_HIDDEN_BUTTON,39,189,30,20,"");
    fl_set_button_shortcut(obj,"#s",1);
    fl_set_object_callback(obj,EnterSquare,1);
  fdui->Square = obj = fl_add_bitmapbutton(FL_NORMAL_BUTTON,39,189,30,20,"");
    fl_set_bitmapbutton_data(obj, square_width, square_height, square_bits);
    fl_set_button_shortcut(obj,"sS",0);
    fl_set_object_callback(obj,EnterSquare,0);

  fdui->NotCmp = obj = fl_add_button(FL_NORMAL_BUTTON,39,212,30,20,"~");
    fl_set_button_shortcut(obj,"~",0);
    fl_set_object_callback(obj,EnterNotCmp,0);

/*
 * HEXADECIMAL DIGITS 
 */
  fdui->HexDigits = fl_bgn_group();
  fdui->DigitA = obj = fl_add_button(FL_TOUCH_BUTTON,74,97,30,20,"A");
    fl_set_button_shortcut(obj,"aA",0);
    fl_set_object_callback(obj,EnterDigit,10);
  fdui->DigitB = obj = fl_add_button(FL_TOUCH_BUTTON,74,120,30,20,"B");
    fl_set_button_shortcut(obj,"bB",0);
    fl_set_object_callback(obj,EnterDigit,11);
  fdui->DigitC = obj = fl_add_button(FL_TOUCH_BUTTON,74,143,30,20,"C");
    fl_set_button_shortcut(obj,"cC",0);
    fl_set_object_callback(obj,EnterDigit,12);
  fdui->DigitD = obj = fl_add_button(FL_TOUCH_BUTTON,74,166,30,20,"D");
    fl_set_button_shortcut(obj,"dD",0);
    fl_set_object_callback(obj,EnterDigit,13);
  fdui->DigitE = obj = fl_add_button(FL_TOUCH_BUTTON,74,189,30,20,"E");
    fl_set_button_shortcut(obj,"eE",0);
    fl_set_object_callback(obj,EnterDigit,14);
  fdui->DigitF = obj = fl_add_button(FL_TOUCH_BUTTON,74,212,30,20,"F");
    fl_set_button_shortcut(obj,"fF",0);
    fl_set_object_callback(obj,EnterDigit,15);
  fl_end_group();


/*
 * BINARY DIGITS 
 */
  fdui->BinDigits = fl_bgn_group();
  fdui->DigitOne = obj = fl_add_button(FL_TOUCH_BUTTON,109,181,30,23,"1");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"1",0);
    fl_set_object_callback(obj,EnterDigit,1);
  fdui->DigitZero = obj = fl_add_button(FL_TOUCH_BUTTON,144,209,30,23,"0");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"0",0);
    fl_set_object_callback(obj,EnterDigit,0);
  fl_end_group();


/*
 * OCTAL DIGITS 
 */
  fdui->OctDigits = obj = fl_bgn_group();
  fdui->DigitSeven = obj = fl_add_button(FL_TOUCH_BUTTON,109,125,30,23,"7");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"7",0);
    fl_set_object_callback(obj,EnterDigit,7);
  fdui->DigitFour = obj = fl_add_button(FL_TOUCH_BUTTON,109,153,30,23,"4");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"4",0);
    fl_set_object_callback(obj,EnterDigit,4);
  fdui->DigitFive = obj = fl_add_button(FL_TOUCH_BUTTON,144,153,30,23,"5");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"5",0);
    fl_set_object_callback(obj,EnterDigit,5);
  fdui->DigitSix = obj = fl_add_button(FL_TOUCH_BUTTON,179,153,30,23,"6");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"6",0);
    fl_set_object_callback(obj,EnterDigit,6);
  fdui->DigitTwo = obj = fl_add_button(FL_TOUCH_BUTTON,144,181,30,23,"2");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"2",0);
    fl_set_object_callback(obj,EnterDigit,2);
  fdui->DigitThree = obj = fl_add_button(FL_TOUCH_BUTTON,179,181,30,23,"3");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"3",0);
    fl_set_object_callback(obj,EnterDigit,3);
  fl_end_group();


/*
 * The comma button gets overlayed by the decimal in dec mode 
 */
  fdui->Comma = obj = fl_add_bitmapbutton(FL_NORMAL_BUTTON,109,209,30,23,"");
    fl_set_bitmapbutton_data(obj, comma_width, comma_height, comma_bits);
    fl_set_button_shortcut(obj,",",0);
    fl_set_object_callback(obj,EnterComma,14);

/*
 * DECIMAL DIGITS 
 */
  fdui->DecDigits = fl_bgn_group();
  fdui->DigitEight = obj = fl_add_button(FL_TOUCH_BUTTON,144,125,30,23,"8");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"8",0);
    fl_set_object_callback(obj,EnterDigit,8);
  fdui->DigitNine = obj = fl_add_button(FL_TOUCH_BUTTON,179,125,30,23,"9");
    fl_set_object_color(obj, FL_SLATEBLUE, FL_BUTTON_COL1);
    fl_set_button_shortcut(obj,"9",0);
    fl_set_object_callback(obj,EnterDigit,9);
  fdui->Decimal = obj = fl_add_bitmapbutton(FL_NORMAL_BUTTON,109,209,30,23,"");
    fl_set_bitmapbutton_data(obj, decimal_width, decimal_height, decimal_bits);
    fl_set_button_shortcut(obj,".,",0);
    fl_set_object_callback(obj,EnterDecimal,0);
  fl_end_group();

/*
 * PERCENT  
 */
  fdui->Percent = obj = fl_add_button(FL_NORMAL_BUTTON,214,209,30,23,"%");
    fl_set_button_shortcut(obj,"%",0);
    fl_set_object_callback(obj,EnterPercent,0);

/*
 * LOGS AND TRIG    
 */
  fdui->Hyper = obj = fl_add_button(FL_PUSH_BUTTON,4,97,30,20,"Hyp");
    fl_set_object_callback(obj,EnterHyper,2);
    fl_set_object_color(obj,FL_BUTTON_COL1, FL_YELLOW);
    fl_set_button_shortcut(obj,"&2",0);

  obj = fl_add_button(FL_HIDDEN_BUTTON,4,120,30,20,"Sin");
    fl_set_object_callback(obj,EnterTrig,0);
    fl_set_button_shortcut(obj,"#&3",0);
  fdui->Sin = obj = fl_add_button(FL_NORMAL_BUTTON,4,120,30,20,"Sin");
    fl_set_object_callback(obj,EnterTrig,0);
    fl_set_button_shortcut(obj,"&3",0);

  obj = fl_add_button(FL_HIDDEN_BUTTON,4,143,30,20,"Cos");
    fl_set_object_callback(obj,EnterTrig,1);
    fl_set_button_shortcut(obj,"#&4",0);
  fdui->Cos = obj = fl_add_button(FL_NORMAL_BUTTON,4,143,30,20,"Cos");
    fl_set_object_callback(obj,EnterTrig,1);
    fl_set_button_shortcut(obj,"&4",0);

  obj = fl_add_button(FL_HIDDEN_BUTTON,4,166,30,20,"Tan");
    fl_set_object_callback(obj,EnterTrig,2);
    fl_set_button_shortcut(obj,"#&5",0);
  fdui->Tan = obj = fl_add_button(FL_NORMAL_BUTTON,4,166,30,20,"Tan");
    fl_set_object_callback(obj,EnterTrig,2);
    fl_set_button_shortcut(obj,"&5",0);

  obj = fl_add_button(FL_HIDDEN_BUTTON,4,189,30,20,"log");
    fl_set_object_callback(obj,EnterLogr,0);
    fl_set_button_shortcut(obj,"#&6",0);
  fdui->Log = obj = fl_add_button(FL_NORMAL_BUTTON,4,189,30,20,"log");
    fl_set_object_callback(obj,EnterLogr,0);
    fl_set_button_shortcut(obj,"&6",0);

  obj = fl_add_button(FL_HIDDEN_BUTTON,4,212,30,20,"logn");
    fl_set_object_callback(obj,EnterLogn,0);
    fl_set_button_shortcut(obj,"#&7",0);
  fdui->Logn = obj = fl_add_button(FL_NORMAL_BUTTON,4,212,30,20,"logn");
    fl_set_object_callback(obj,EnterLogn,0);
    fl_set_button_shortcut(obj,"&7",0);

/*
 * THE DISPLAY AND ITS HIDDEN QUIT, CUT, PASTE AND HELP BUTTONS
 */
  fdui->CalcDisplay = obj = fl_add_text(FL_NORMAL_TEXT,3,30,311,25,"");
    fl_set_object_boxtype(obj,FL_DOWN_BOX);
    fl_set_object_lsize(obj,FL_LARGE_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT);
    fl_set_object_lstyle(obj,FL_TIMESITALIC_STYLE);
  obj = fl_add_button(FL_HIDDEN_BUTTON,3,28,1,1,"");
    fl_set_button_shortcut(obj,"^c",0);
	fl_set_object_callback(obj,CutPasteDisplay,1);
  obj = fl_add_button(FL_HIDDEN_BUTTON,5,28,1,1,"");
    fl_set_button_shortcut(obj,"^v",0);
	fl_set_object_callback(obj,CutPasteDisplay,2);
  obj = fl_add_button(FL_HIDDEN_BUTTON,7,28,1,1,"");
    fl_set_button_shortcut(obj,"&1",0);
	fl_set_object_callback(obj,CalcHelp,2);
  obj = fl_add_button(FL_HIDDEN_BUTTON,9,28,1,1,"");
    fl_set_button_shortcut(obj,"&12",0);
	fl_set_object_callback(obj,ExitProgram,0);

  fl_end_form();

  return fdui;
}
/*---------------------------------------*/

char *build_constants()
{
/*
 * Build the constants menu string and the constants array from the data
 * found in xfrmcalc.constants file in the /etc/xfrmcalc directory. Each
 * line in that file contains two ':' separated fields. The first is the
 * string that is to into the menu. The second is the scanf-able float value 
 * of the constant value itself.
 */
	FILE		*consts;
	char		*menu_items = NULL, con_name[BUF_SIZE], *temp, *val_str;
	CALCAMNT	con_value, *temp_value;
	int			mi_size = 0, mi_left = 0, num_cons = 0, cons_left = 0, i;

	if ((consts = fopen(XCONS,"r")) == NULL)
		return NULL;

	while(fgets(con_name, BUF_SIZE-1, consts) != NULL)
	{
		if ((val_str = strchr(con_name, ':')) != NULL)
		{
			*val_str = '\0';
			if (!sscanf(++val_str, CALCFRMT, &con_value))
				continue;

			if (mi_left < (strlen(con_name)+1))
			{
				temp = (char *)calloc(mi_size + BUF_SIZE, 1);
				if (mi_size > 0) {
					strcpy(temp, menu_items);
					free(menu_items);
				}
				mi_left  = BUF_SIZE;
				mi_size += BUF_SIZE;
				menu_items = temp;
			}

			if (cons_left < 1)
			{
				temp_value = (CALCAMNT *)calloc(num_cons+5, sizeof(CALCAMNT));

				if (num_cons > 0)
				{
					for (i= 0; i < num_cons; ++i)
						temp_value[i] = constants[i];
					free(constants);
				}
				cons_left = 5;
				constants = temp_value;
			}

			if (num_cons > 0)
			{
				strcat(menu_items,"|");
				mi_left--;
			}
			strcat(menu_items, con_name);
			mi_left -= strlen(con_name);
			constants[num_cons++] = con_value;
			cons_left--;
		}

	}
	return menu_items;
}

char * build_ext_funcs()
{
/*
 * Build the functions menu string and the functions array from the *.cal
 * files found in the /etc/xfrmcalc directory. The second line of each file,
 * must contain two ':' separated fields. The first is the string that is to 
 * go into the menu. The second is the scanf-able int value detailing how many 
 * arguments the function takes. -1 means the function swallows all arguments
 * on the stack. The line must also be prefixed by a '#' characetr in order
 * for perl to accept it!
 */
	FILE		*prog;
	FL_Dirlist 	*func_progs;
	int			mi_size = 0, mi_left = 0, num_progs = 0, i, j = 0, num_args;
	char		*menu_items = NULL, file_name[BUF_SIZE], argc_str[10],
				prog_name[BUF_SIZE], *temp;

	func_progs = (FL_Dirlist *)fl_get_dirlist(XCALC, "*.cal", &num_progs, 0);

	for (i = 0; i < num_progs; ++i)
	{
		if (i == 0)
			ext_funcs = (ext_func*)calloc(num_progs, sizeof(ext_func));
	
		strcpy(file_name, XCALC);
		strcat(file_name,func_progs[i].name);
		if ((prog = fopen(file_name, "r")) != NULL
		&& fgets(prog_name, BUF_SIZE-1, prog) != NULL /* DELIBERATELY TWO */
		&& fgets(prog_name, BUF_SIZE-1, prog) != NULL /* we need line 2   */
		&& (temp = strchr(prog_name, ':')) != NULL
		&& (sscanf(&temp[1],"%i",&num_args)) > 0)
		{
			*temp = '\0';
			if (num_args < 0)
				strcpy(argc_str, " (V)");
			else
				sprintf(argc_str," (%i)", num_args);
			if (mi_left < (strlen(&prog_name[1])+strlen(argc_str)+1))
			{
				temp = (char *)calloc(mi_size + BUF_SIZE, 1);
				if (mi_size > 0) {
					strcpy(temp, menu_items);
					free(menu_items);
				}
				mi_left  = BUF_SIZE;
				mi_size += BUF_SIZE;
				menu_items = temp;
			}
			if (j > 0)
			{
				strcat(menu_items,"|");
				mi_left--;
			}
			strcat(menu_items, &prog_name[1]);
			strcat(menu_items, argc_str);
			mi_left -= strlen(&prog_name[1]);
			ext_funcs[j].name = (char *)malloc(strlen(file_name)+1);
			strcpy(ext_funcs[j].name, file_name);
			ext_funcs[j].args = num_args;
			j++;
		}
		if (prog != NULL)
			fclose(prog);
				
	}
	if (func_progs != NULL)
		fl_free_dirlist(func_progs);

	return(menu_items);
}
